from lib.fileutils import *
from dataclasses import dataclass, field
from typing import List, Union
import math
import prettyprinter
from prettyprinter import pprint


prettyprinter.install_extras()


NAM_FlagHasHandles = (1 << 0)
NAM_FlagCaseInsensitive = (1 << 1)
ANM_NODE_FLAG_POS_ANIMATED = (1 << 0)
ANM_NODE_FLAG_ROT_ANIMATED = (1 << 1)
ANM_NODE_FLAG_ROL_ANIMATED = (1 << 2)
ANM_NODE_FLAG_SCL_ANIMATED = (1 << 3)
ANM_NODE_FLAG_AFF_ANIMATED = (1 << 4)
ANM_NODE_FLAG_USR_ANIMATED = (1 << 5)
ANM_NODE_FLAG_ANIMATED = ((1 << 6) - 1)


@dataclass
class Vec3:
    x: float
    y: float
    z: float

    def distance(self, other: 'Vec3') -> float:
        return math.sqrt(math.pow(self.x - other.x, 2) + math.pow(self.y - other.y, 2) + math.pow(self.z - other.z, 2))

    @classmethod
    def create(cls, b, offset=None):
        old = b.tell()
        if offset:
            b.seek(offset)
        x = read_float_and_print("Vec3->x", b)
        y = read_float_and_print("Vec3->y", b)
        z = read_float_and_print("Vec3->z", b)
        if offset:
            b.seek(old)
        return cls(x=x, y=y, z=z)


@dataclass
class Vec3Indexed(Vec3):
    index: int

    @classmethod
    def create(cls, b, offset=None):
        old = b.tell()
        if offset:
            b.seek(offset)
        x = read_float_and_print("Vec3->x", b)
        y = read_float_and_print("Vec3->y", b)
        z = read_float_and_print("Vec3->z", b)
        index = read_int_and_print("Vec3->index", b)
        if offset:
            b.seek(old)
        return cls(x=x, y=y, z=z, index=index)


@dataclass
class Vec4(Vec3):
    w: float

    @classmethod
    def create(cls, b, offset=None):
        old = b.tell()
        if offset:
            b.seek(offset)
        x = read_float_and_print("Vec4->x", b)
        y = read_float_and_print("Vec4->y", b)
        z = read_float_and_print("Vec4->z", b)
        w = read_float_and_print("Vec4->w", b)
        if offset:
            b.seek(old)
        return cls(x=x, y=y, z=z, w=w)


@dataclass
class AnmObj:
    custom: int = field(repr=False)
    skin: int = field(repr=False)
    zero_cluster: int
    total_number_vertices: int
    num_clusters: int
    clusters: List['AnmCluster']

    @classmethod
    def create(cls, b, offset=None):
        old = b.tell()
        if offset:
            b.seek(offset)
        custom = read_int_and_print("ANM_Obj->Custom", b)
        skin = read_int_and_print("ANM_Obj->Skin", b)
        zero = read_int_and_print("ANM_Obj->ZeroCluster", b)
        total_vertices = read_int_and_print("ANM_Obj->TotalNumberOfVertices", b)
        num_clusters = read_int_and_print("ANM_Obj->NumberOfClusters", b)
        clusters_offset = read_int_and_print("ANM_Obj->Clusters", b)
        clusters = []
        _old1 = b.tell()
        b.seek(clusters_offset)
        for _ in range(num_clusters):
            cluster = AnmCluster.create(b)
            clusters.append(cluster)

        b.seek(_old1)
        if offset:
            b.seek(old)
        r = cls(custom=custom, skin=skin, zero_cluster=zero, total_number_vertices=total_vertices, num_clusters=num_clusters, clusters=clusters)
        for cluster in clusters:
            cluster.obj = r
        return r


@dataclass
class AnmCluster:
    custom: int = field(repr=False)
    obj: Union[AnmObj, None] = field(repr=False)
    handle: int = field(repr=False)
    bounding_box: List[Vec3] = field(repr=False)  # size 2
    num_vertices: int
    vertices: List[Vec3Indexed] = field(repr=False)
    bone_name: str = ""

    @classmethod
    def create(cls, b, offset=None):
        old = b.tell()
        if offset:
            b.seek(offset)
        custom = read_int_and_print("ANM_Cluster->Custom", b)
        obj_offset = read_int_and_print("ANM_Cluster->Obj", b)
        if obj_offset == 0:  # ugly hack, dunno why we need that
            _ = read_int_and_print("ANM_Cluster->Custom", b)
            _ = read_int_and_print("ANM_Cluster->Obj", b)
        obj = None  # is set by parent Obj
        handle = read_int_and_print("ANM_Cluster->Handle", b)
        bbox1 = Vec3.create(b)
        bbox2 = Vec3.create(b)
        bounding_box = [bbox1, bbox2]
        num_vertices = read_int_and_print("ANM_Cluster->num_vertices", b)
        vertices = []
        for _ in range(num_vertices):
            v = Vec3Indexed.create(b)
            vertices.append(v)

        if offset:
            b.seek(old)
        return cls(custom=custom, obj=obj, handle=handle, bounding_box=bounding_box, num_vertices=num_vertices, vertices=vertices)


class NamDictionnary:
    @staticmethod
    def create(b, cls, depth=1, offset=None):
        r = {}

        old = b.tell()
        if offset:
            b.seek(offset)

        read_int_and_print("NAM_Dictionnary->pMemGroup", b)
        namemapper = read_int_and_print("NAM_Dictionnary->pNameMapper", b)
        flags = read_int_and_print("NAM_Dictionnary->Flags", b)
        print("Flags: NAM_FlagHasHandles: %s, NAM_FlagCaseInsensitive: %s" % (flags & NAM_FlagHasHandles, flags & NAM_FlagCaseInsensitive))
        read_short_and_print("NAM_Dictionnary->NameHandle", b)

        b.seek(namemapper)
        read_int_and_print("pMemGroup", b)
        read_int_and_print("MemBaseSize", b)
        read_int_and_print("MemBlocks", b)
        read_int_and_print("MemSize", b)
        read_int_and_print("MemFree", b)
        read_int_and_print("pMem", b)
        read_int_and_print("MaxMappings", b)
        totalmappings = read_int_and_print("TotalMappings", b)
        read_int_and_print("TotalKeys", b)
        p_key_contexts = read_int_and_print("pKeyContexts", b)

        print("MAP_Image->ANM_Skin->NAM_Dictionnary->MAP_Mapper->MAP_KeyContext")
        b.seek(p_key_contexts)
        print("\tMAP_Image->ANM_Skin->NAM_Dictionnary->MAP_Mapper->MAP_KeyContext->MAP_BinSearchContext")
        _ = read_int_and_print("\tSize", b)
        read_int_and_print("\tpCallback", b)
        read_int_and_print("\tIndex", b)
        read_int_and_print("\tTest", b)
        p_lut = read_int_and_print("pLUT", b)
        read_int_and_print("Reserve", b)
        read_int_and_print("pData", b)

        for i in range(totalmappings):
            b.seek(p_lut)
            pindx = read_int(b)
            for _ in range(i):
                pindx = read_int(b)
            print("LUT INDEX: %s" % pindx)
            b.seek(pindx)
            if not flags & NAM_FlagHasHandles:
                read_short_and_print("pPtr", b)
                read_short_and_print("pPtr2", b)
                val = read_int_and_print("Value", b)
                name = read_string_until_none_and_print("name", b)
                if depth > 1:
                    item = NamDictionnary.create(b, cls, depth-1, val)
                else:
                    item = cls.create(b, val)
                r[name] = item
            else:
                refcnt = read_short_and_print("RefCnt", b)
                handle = read_short_and_print("Handle", b)
                read_int_and_print("Test", b)
                read_int_and_print("Test", b)
                bone_name = read_string_until_none_and_print("BoneName", b)
                r[str(i)] = {"BoneName": bone_name}

        if offset:
            b.seek(old)

        return r


@dataclass
class GeoMat4x4:
    m: List[float]

    @classmethod
    def create(cls, b, offset=None):
        old = b.tell()
        if offset:
            b.seek(offset)

        m = []
        for _ in range(4):
            for _j in range(4):
                m.append(read_float(b))

        r = cls(m=m)

        if offset:
            b.seek(old)

        return r


@dataclass
class GeoMat:
    mat_class: int
    s: GeoMat4x4

    @classmethod
    def create(cls, b, offset=None):
        old = b.tell()
        if offset:
            b.seek(offset)

        mat_class = read_int(b)
        s = GeoMat4x4.create(b)

        r = cls(mat_class=mat_class, s=s)

        if offset:
            b.seek(old)

        return r


@dataclass
class GeoAffine:
    transform: Vec3
    quat: Vec4
    scale: Vec3
    affine_class: int

    @classmethod
    def create(cls, b, offset=None):
        old = b.tell()
        if offset:
            b.seek(offset)

        trans = Vec3.create(b)
        quat = Vec4.create(b)
        scale = Vec3.create(b)
        affine_class = read_int(b)

        r = cls(transform=trans, quat=quat, scale=scale, affine_class=affine_class)

        if offset:
            b.seek(old)

        return r


@dataclass
class AnmTransform:
    time: int
    affine: GeoAffine

    @classmethod
    def create(cls, b, offset=None):
        old = b.tell()
        if offset:
            b.seek(offset)

        time = read_int(b)
        affine = GeoAffine.create(b)

        r = cls(time=time, affine=affine)

        if offset:
            b.seek(old)

        return r


@dataclass
class AnmNode:
    custom: int
    node_type: int
    x_form_type: int
    flags: int
    p_anim: int
    p_parent: int
    p_target: int
    p_usercall: int
    p_usercontext: int
    stamp: int
    geo_mat: GeoMat
    transform: AnmTransform
    pos_keys: int
    rot_keys: int
    scl_keys: int

    @classmethod
    def create(cls, b, offset=None):
        old = b.tell()
        if offset:
            b.seek(offset)

        custom = read_int_and_print("ANM_Node->Custom", b)
        node_type = read_short_and_print("ANM_Node->Type", b)
        x_form_type = read_short_and_print("ANM_Node->XformType", b)
        flags = read_int_and_print("ANM_Node->Flags", b)
        p_anim = read_int_and_print("ANM_Node->pAnim", b)
        p_parent = read_int_and_print("ANM_Node->pParent", b)
        p_target = read_int_and_print("ANM_Node->pTarget", b)
        p_usercall = read_int_and_print("ANM_Node->pUserCall", b)
        p_usercontext = read_int_and_print("ANM_Node->pUserContext", b)
        stamp = read_int_and_print("ANM_Node->Stamp", b)
        geo_mat = GeoMat.create(b)
        transform = AnmTransform.create(b)
        pos_keys = read_int_and_print("ANM_Node->pPosKeys", b)
        rot_keys = read_int_and_print("ANM_Node->pRotKeys", b)
        scl_keys = read_int_and_print("ANM_Node->pSclKeys", b)

        """
        if nodeflags & ANM_NODE_FLAG_POS_ANIMATED:
            b.seek(posKeys)
            read_byte_and_print("\t\tType", b)
            read_byte_and_print("\t\tORTBefore", b)
            read_byte_and_print("\t\tORTAfter", b)
            read_byte_and_print("\t\tFlags", b)
            read_int_and_print("\t\tpInterpolate", b)
            read_int_and_print("\t\tNumberOfKeys", b)
            read_int_and_print("\t\tpKeys", b)
        """

        r = cls(custom=custom, node_type=node_type, x_form_type=x_form_type, flags=flags, p_anim=p_anim,
                p_parent=p_parent, p_target=p_target, p_usercall=p_usercall, p_usercontext=p_usercontext,
                stamp=stamp, geo_mat=geo_mat, transform=transform, pos_keys=pos_keys, rot_keys=rot_keys,
                scl_keys=scl_keys)

        if offset:
            b.seek(old)

        return r


@dataclass
class AnmAnim:
    custom: int
    fileversion: int
    image: int
    memgroup: int
    start_time: int
    end_time: int
    ticks_per_frame: int
    framerate: int
    tree_dictionnary: dict
    default_obj_dictionnary: dict
    num_nodes: int

    extra_stamp: int
    extra_flags: int

    extra_time1: int
    extra_stamp1: int

    extra_time2: int
    extra_stamp2: int

    extra_factor: float
    extra_other_anim: int
    name: str

    @classmethod
    def create(cls, b, offset=None):
        old = b.tell()
        if offset:
            b.seek(offset)

        custom = read_int_and_print("ANM_Anim->Custom", b)
        version = read_int_and_print("ANM_Anim->FileVersion", b)
        image = read_int_and_print("ANM_Anim->pImage", b)
        memgroup = read_int_and_print("ANM_Anim->pMemGroup", b)

        start_time = read_int_and_print("ANM_Anim->StartTime", b)
        end_time = read_int_and_print("ANM_Anim->EndTime", b)
        ticks_per_frame = read_int_and_print("TicksPerFrame", b)
        framerate = read_int_and_print("FrameRate", b)
        treedict = read_int_and_print("pTreeDictionary", b)
        defaultobjdict = read_int_and_print("pDefaultObjDictionary", b)
        num_nodes = read_int_and_print("NumberOfNodes", b)
        stamp = read_int_and_print("ANM_AnimExtra->Stamp", b)
        flags = read_int_and_print("ANM_AnimExtra->Flags", b)
        time1 = read_int_and_print("ANM_AnimExtra->Time1", b)
        stamp1 = read_int_and_print("ANM_AnimExtra->Stamp1", b)
        time2 = read_int_and_print("ANM_AnimExtra->Time2", b)
        stamp2 = read_int_and_print("ANM_AnimExtra->Stamp2", b)
        factor = read_float_and_print("ANM_AnimExtra->Factor", b)
        otheranim = read_int_and_print("ANM_AnimExtra->OtherAnim", b)
        name = read_string_until_none(b)

        treedict = NamDictionnary.create(b, AnmNode, depth=2, offset=treedict)
        defaultobjdict = NamDictionnary.create(b, AnmNode, offset=defaultobjdict)

        r = cls(custom=custom, fileversion=version, image=image, memgroup=memgroup, start_time=start_time,
                end_time=end_time, ticks_per_frame=ticks_per_frame, framerate=framerate, tree_dictionnary=treedict,
                default_obj_dictionnary=defaultobjdict, num_nodes=num_nodes, extra_stamp=stamp, extra_flags=flags,
                extra_time1=time1, extra_stamp1=stamp1, extra_time2=time2, extra_stamp2=stamp2, extra_factor=factor,
                extra_other_anim=otheranim, name=name)

        if offset:
            b.seek(old)

        return r


@dataclass
class AnmSkin:
    custom: int
    file_version: int
    image: int
    memgroup: int
    node_dictionnary: dict
    tree_dictionnary: dict
    default_obj_dictionnary: dict
    total_number_trees: int
    total_number_objs: int
    total_number_clusters: int
    name: str

    @staticmethod
    def link(nodedict, defaultobjtree):
        for obj_name, obj in defaultobjtree.items():
            for cluster in obj.clusters:
                cluster_handle = cluster.handle
                k = str(cluster_handle - 1)
                cluster.bone_name = nodedict[k]["BoneName"]

    @classmethod
    def create(cls, b, offset=None):
        old = b.tell()
        if offset:
            b.seek(offset)

        custom = read_int_and_print("pCustom", b)
        version = read_int_and_print("FileVersion", b)
        image = read_int_and_print("pImage", b)
        memgroup = read_int_and_print("pMemGroup", b)
        node = read_int_and_print("pNodeDictionary", b)
        tree = read_int_and_print("pTreeDictionary", b)
        defaultobj = read_int_and_print("pDefaultObjDictionary", b)
        num_trees = read_int_and_print("TotalNumberOfTrees", b)
        num_objs = read_int_and_print("TotalNumberOfObjs", b)
        num_clusters = read_int_and_print("TotalNumberOfClusters", b)
        name = read_string_until_none_and_print("Name", b)

        nodedict = NamDictionnary.create(b, AnmObj, offset=node)
        treedict = NamDictionnary.create(b, AnmObj, depth=2, offset=tree)
        defaultobjdict = NamDictionnary.create(b, AnmObj, offset=defaultobj)

        # AnmSkin.link(nodedict, defaultobjdict)
        r = cls(custom=custom, file_version=version, image=image, memgroup=memgroup, node_dictionnary=nodedict, tree_dictionnary=treedict, default_obj_dictionnary=defaultobjdict, total_number_trees=num_trees, total_number_objs=num_objs, total_number_clusters=num_clusters, name=name)

        if offset:
            b.seek(old)

        return r


@dataclass
class AnmFile:
    @staticmethod
    def parse(b, cls):
        _ = read_int_and_print("size", b)
        memgroupoffset = read_int_and_print("memgroupoffset", b)
        b.seek(memgroupoffset)
        read_int_and_print("\tNode", b)
        read_int_and_print("\tSize", b)
        read_int_and_print("\tFlags", b)

        read_int_and_print("pMemGroup", b)
        read_int_and_print("MemList", b)
        read_int_and_print("TotalSize", b)
        read_int_and_print("TotalBlocks", b)
        base = read_int_and_print("pBase", b)
        return cls.create(b, base)
